@@ -34,7 +34,7 @@ use parking_lot::Mutex;
 use rand::Rng;
 use std::{
     cmp::{max, Eq, PartialEq},
-    collections::{hash_map::Entry, HashMap, VecDeque},
+    collections::{hash_map::Entry, HashMap, HashSet, VecDeque},
     error, fmt,
     hash::Hash,
     iter, mem,
@@ -48,6 +48,72 @@ use tokio::{
     task,
     time::{self, Instant},
 };
+use triomphe::Arc as TArc;
+
+type StreamsInner<T> = Arc<Vec<(T, ChanWrap<Pooled<Vec<(SubId, Event)>>>)>>;
+
+lazy_static! {
+    static ref HCSTREAMS: Mutex<HashSet<StreamsInner<ChanId>>> = Mutex::new(HashSet::new());
+    static ref HCDVSTREAMS: Mutex<HashSet<StreamsInner<UpdatesFlags>>> =
+        Mutex::new(HashSet::new());
+    static ref BATCHES: Pool<Vec<(SubId, Event)>> = Pool::new(64, 16384);
+    static ref DECODE_BATCHES: Pool<Vec<From>> = Pool::new(64, 16384);
+}
+
+macro_rules! hcstreams {
+    ($name:ident, $static:ident, $typ:ty) => {
+        #[derive(Debug)]
+        struct $name(StreamsInner<$typ>);
+
+        impl $name {
+            fn new() -> $name {
+                let mut inner = $static.lock();
+                match inner.get(&vec![]) {
+                    Some(empty) => $name(empty.clone()),
+                    None => {
+                        let r = Arc::new(vec![]);
+                        inner.insert(r.clone());
+                        $name(r)
+                    }
+                }
+            }
+
+            fn add(
+                &self,
+                chanid: $typ,
+                chan: ChanWrap<Pooled<Vec<(SubId, Event)>>>,
+            ) -> $name {
+                let mut dead = false;
+                let mut vec = Vec::clone(&self.0);
+                vec.retain(|(_, c)| {
+                    if c.0.is_closed() {
+                        dead = true;
+                        false
+                    } else {
+                        true
+                    }
+                });
+                vec.push((chanid, chan));
+                vec.sort_by_key(|(id, _)| *id);
+                let mut inner = $static.lock();
+                if dead {
+                    inner.remove(&self.0);
+                }
+                match inner.get(&vec) {
+                    Some(cur) => $name(Arc::clone(cur)),
+                    None => {
+                        let t = Arc::new(vec);
+                        inner.insert(t.clone());
+                        $name(t)
+                    }
+                }
+            }
+        }
+    };
+}
+
+hcstreams!(Streams, HCSTREAMS, ChanId);
+hcstreams!(DvStreams, HCDVSTREAMS, UpdatesFlags);
 
 #[derive(Debug)]
 pub struct PermissionDenied;
@@ -72,6 +138,8 @@ impl fmt::Display for NoSuchValue {
 impl error::Error for NoSuchValue {}
 
 atomic_id!(SubId);
+atomic_id!(SubscriberId);
+atomic_id!(ConId);
 
 bitflags! {
     pub struct SubscribeFlags: u32 {
@@ -112,6 +180,7 @@ bitflags! {
 #[derive(Debug)]
 struct SubscribeValRequest {
     path: Path,
+    sub_id: SubId,
     timestamp: u64,
     permissions: u32,
     token: Bytes,
@@ -170,9 +239,9 @@ impl Pack for Event {
 struct ValInner {
     sub_id: SubId,
     id: Id,
-    addr: SocketAddr,
+    conid: ConId,
     connection: BatchSender<ToCon>,
-    last: Arc<Mutex<Event>>,
+    last: TArc<Mutex<Event>>,
 }
 
 impl Drop for ValInner {
@@ -272,7 +341,7 @@ enum DvState {
 struct DvalInner {
     sub_id: SubId,
     sub: DvState,
-    streams: Vec<(UpdatesFlags, Sender<Pooled<Vec<(SubId, Event)>>>)>,
+    streams: DvStreams,
 }
 
 #[derive(Debug, Clone)]
@@ -350,9 +419,9 @@ impl Dval {
         tx: mpsc::Sender<Pooled<Vec<(SubId, Event)>>>,
     ) {
         let mut t = self.0.lock();
-        t.streams.retain(|(_, c)| !c.is_closed());
-        if !t.streams.iter().any(|(_, s)| tx.same_receiver(s)) {
-            t.streams.push((flags, tx.clone()));
+        let c = ChanWrap(tx.clone());
+        if !t.streams.0.iter().any(|(_, s)| &c == s) {
+            t.streams = t.streams.add(flags, c);
         }
         if let DvState::Subscribed(ref sub) = t.sub {
             let m = ToCon::Stream { tx, sub_id: t.sub_id, id: sub.0.id, flags };
@@ -479,6 +548,7 @@ fn pick(n: usize) -> usize {
 
 #[derive(Debug)]
 struct SubscriberInner {
+    id: SubscriberId,
     resolver: ResolverRead,
     connections: HashMap<SocketAddr, BatchSender<ToCon>, FxBuildHasher>,
     recently_failed: HashMap<SocketAddr, Instant, FxBuildHasher>,
@@ -491,6 +561,15 @@ struct SubscriberInner {
 }
 
 impl SubscriberInner {
+    fn durable_id(&self, path: &Path) -> Option<SubId> {
+        self.durable_dead
+            .get(path)
+            .or_else(|| self.durable_pending.get(path))
+            .or_else(|| self.durable_alive.get(path))
+            .and_then(|w| w.upgrade())
+            .map(|d| d.id())
+    }
+
     fn choose_addr(&mut self, resolved: &Resolved) -> (SocketAddr, Bytes) {
         use rand::seq::IteratorRandom;
         let flags = unsafe { PublishFlags::from_bits_unchecked(resolved.flags) };
@@ -528,6 +607,13 @@ impl SubscriberWeak {
     }
 }
 
+#[derive(Debug, Clone, Copy)]
+pub struct DurableStats {
+    pub alive: usize,
+    pub pending: usize,
+    pub dead: usize,
+}
+
 /// create subscriptions
 #[derive(Clone, Debug)]
 pub struct Subscriber(Arc<Mutex<SubscriberInner>>);
@@ -538,6 +624,7 @@ impl Subscriber {
         let (tx, rx) = mpsc::unbounded();
         let resolver = ResolverRead::new(resolver, desired_auth.clone());
         let t = Subscriber(Arc::new(Mutex::new(SubscriberInner {
+            id: SubscriberId::new(),
             resolver,
             desired_auth,
             connections: HashMap::with_hasher(FxBuildHasher::default()),
@@ -552,6 +639,23 @@ impl Subscriber {
         Ok(t)
     }
 
+    /// Return a unique identifier for this subscriber instance. The
+    /// identifier will be unique across all subscribers created in
+    /// this process, but not across processes or machines.
+    pub fn id(&self) -> SubscriberId {
+        self.0.lock().id
+    }
+
+    /// return stats about durable subscriptions
+    pub fn durable_stats(&self) -> DurableStats {
+        let t = self.0.lock();
+        DurableStats {
+            alive: t.durable_alive.len(),
+            pending: t.durable_pending.len(),
+            dead: t.durable_dead.len(),
+        }
+    }
+
     pub fn resolver(&self) -> ResolverRead {
         self.0.lock().resolver.clone()
     }
@@ -568,25 +672,31 @@ impl Subscriber {
             }
         }
         fn update_retry(subscriber: &mut SubscriberInner, retry: &mut Option<Instant>) {
-            *retry = subscriber
-                .durable_dead
-                .values()
-                .filter_map(|w| w.upgrade())
-                .map(|ds| match &ds.0.lock().sub {
-                    DvState::Dead(dead) => dead.next_try,
-                    DvState::Subscribed(_) => unreachable!(),
-                })
-                .fold(None, |min, v| match min {
-                    None => Some(v),
-                    Some(min) => {
-                        if v < min {
-                            Some(v)
-                        } else {
-                            Some(min)
+            let now = Instant::now();
+            task::block_in_place(|| {
+                *retry = None;
+                for w in subscriber.durable_dead.values() {
+                    if let Some(dv) = w.upgrade() {
+                        let next_try = match &dv.0.lock().sub {
+                            DvState::Dead(dead) => dead.next_try,
+                            DvState::Subscribed(_) => unreachable!(),
+                        };
+                        match retry {
+                            None => {
+                                *retry = Some(next_try);
+                            }
+                            Some(retry) => {
+                                if next_try < *retry {
+                                    *retry = next_try;
+                                }
+                            }
+                        }
+                        if next_try <= now {
+                            break;
                         }
                     }
-                })
-                .map(|t| t + Duration::from_secs(1));
+                }
+            })
         }
         async fn do_resub(
             subscriber: &SubscriberWeak,
@@ -596,31 +706,43 @@ impl Subscriber {
             info!("doing resubscriptions");
             let now = Instant::now();
             let (batch, timeout) = {
+                let mut dead = Vec::new();
                 let mut batch = Vec::new();
                 let mut subscriber = subscriber.0.lock();
                 let subscriber = &mut *subscriber;
                 let durable_dead = &mut subscriber.durable_dead;
                 let durable_pending = &mut subscriber.durable_pending;
                 let mut max_tries = 1;
-                durable_dead.retain(|p, w| match w.upgrade() {
-                    None => false,
-                    Some(s) => {
-                        let (next_try, tries) = {
-                            let dv = s.0.lock();
-                            match &dv.sub {
-                                DvState::Dead(d) => (d.next_try, d.tries),
-                                DvState::Subscribed(_) => unreachable!(),
+                let mut total_retries = 0;
+                task::block_in_place(|| {
+                    for (p, w) in durable_dead.iter() {
+                        match w.upgrade() {
+                            None => {
+                                dead.push(p.clone());
+                            }
+                            Some(s) => {
+                                let (next_try, tries) = {
+                                    let mut dv = s.0.lock();
+                                    match &mut dv.sub {
+                                        DvState::Dead(d) => (d.next_try, d.tries),
+                                        DvState::Subscribed(_) => unreachable!(),
+                                    }
+                                };
+                                if next_try <= now {
+                                    batch.push(p.clone());
+                                    durable_pending.insert(p.clone(), w.clone());
+                                    max_tries = max(max_tries, tries);
+                                    total_retries += 1;
+                                    if total_retries >= 100_000 {
+                                        break;
+                                    }
+                                }
                             }
-                        };
-                        if next_try > now {
-                            true
-                        } else {
-                            batch.push(p.clone());
-                            durable_pending.insert(p.clone(), w.clone());
-                            max_tries = max(max_tries, tries);
-                            false
                         }
                     }
+                    for p in dead.iter().chain(batch.iter()) {
+                        durable_dead.remove(p);
+                    }
                 });
                 let timeout = 30 + max(10, batch.len() / 10000) * max_tries;
                 (batch, Duration::from_secs(timeout as u64))
@@ -665,10 +787,9 @@ impl Subscriber {
                             },
                             Ok(sub) => {
                                 info!("resubscription success {}", p);
-                                dv.streams.retain(|(_, c)| !c.is_closed());
-                                for (flags, tx) in dv.streams.iter().cloned() {
+                                for (flags, tx) in dv.streams.0.iter().cloned() {
                                     sub.0.connection.send(ToCon::Stream {
-                                        tx,
+                                        tx: tx.0,
                                         sub_id: dv.sub_id,
                                         id: sub.0.id,
                                         flags: flags | UpdatesFlags::BEGIN_WITH_LAST,
@@ -722,17 +843,12 @@ impl Subscriber {
         }
         let subscriber = self.downgrade();
         task::spawn(async move {
-            let mut incoming = Batched::new(incoming.fuse(), 100_000);
+            let mut incoming = Batched::new(incoming.fuse(), 1_000_000_000);
             let mut subscriptions = VecDeque::new();
             let mut subscription_batch = Vec::new();
             let mut retry: Option<Instant> = None;
             loop {
                 select_biased! {
-                    _ = wait_retry(retry).fuse() => {
-                        if let Some(set) = do_resub(&subscriber, &mut retry).await {
-                            subscriptions.push_back(Batched::new(set, 100_000));
-                        }
-                    },
                     m = incoming.next() => match m {
                         None => break,
                         Some(BatchItem::InBatch(())) => (),
@@ -750,8 +866,20 @@ impl Subscriber {
                                 &mut subscription_batch,
                                 &mut retry
                             );
+                            if let Some(t) = retry {
+                                if Instant::now() >= t {
+                                    if let Some(set) = do_resub(&subscriber, &mut retry).await {
+                                        subscriptions.push_back(Batched::new(set, 100_000));
+                                    }
+                                }
+                            }
                         }
-                    }
+                    },
+                    _ = wait_retry(retry).fuse() => {
+                        if let Some(set) = do_resub(&subscriber, &mut retry).await {
+                            subscriptions.push_back(Batched::new(set, 100_000));
+                        }
+                    },
                 }
             }
         });
@@ -863,6 +991,7 @@ impl Subscriber {
                             pending.insert(p, St::Error(anyhow!("path not found")));
                         } else {
                             let addr = t.choose_addr(&resolved);
+                            let sub_id = t.durable_id(&p).unwrap_or_else(SubId::new);
                             let con = t.connections.entry(addr.0).or_insert_with(|| {
                                 let (tx, rx) = batch_channel::channel();
                                 let target_spn = match resolved.krb5_spns.get(&addr.0) {
@@ -907,6 +1036,7 @@ impl Subscriber {
                             let con_ = con.clone();
                             let r = con.send(ToCon::Subscribe(SubscribeValRequest {
                                 path: p.clone(),
+                                sub_id,
                                 timestamp: resolved.timestamp,
                                 permissions: resolved.permissions as u32,
                                 token: addr.1,
@@ -1035,7 +1165,7 @@ impl Subscriber {
                 tries: 0,
                 next_try: Instant::now(),
             })),
-            streams: Vec::new(),
+            streams: DvStreams::new(),
         })));
         t.durable_dead.insert(path, s.downgrade());
         let _ = t.trigger_resub.unbounded_send(());
@@ -1067,14 +1197,14 @@ impl Subscriber {
 
 struct Sub {
     path: Path,
-    streams: Vec<(SubId, ChanId, Sender<Pooled<Vec<(SubId, Event)>>>)>,
-    last: Arc<Mutex<Event>>,
-    collect_last: bool,
+    sub_id: SubId,
+    streams: Streams,
+    last: Option<TArc<Mutex<Event>>>,
 }
 
 type ByChan = HashMap<
     ChanId,
-    (Sender<Pooled<Vec<(SubId, Event)>>>, Pooled<Vec<(SubId, Event)>>),
+    (ChanWrap<Pooled<Vec<(SubId, Event)>>>, Pooled<Vec<(SubId, Event)>>),
     FxBuildHasher,
 >;
 
@@ -1083,16 +1213,18 @@ fn unsubscribe(
     by_chan: &mut ByChan,
     sub: Sub,
     id: Id,
-    addr: SocketAddr,
+    conid: ConId,
 ) {
-    for (sub_id, chan_id, c) in sub.streams.iter() {
+    for (chan_id, c) in sub.streams.0.iter() {
         by_chan
             .entry(*chan_id)
             .or_insert_with(|| (c.clone(), BATCHES.take()))
             .1
-            .push((*sub_id, Event::Unsubscribed))
+            .push((sub.sub_id, Event::Unsubscribed))
+    }
+    if let Some(last) = &sub.last {
+        *last.lock() = Event::Unsubscribed;
     }
-    *sub.last.lock() = Event::Unsubscribed;
     if let Some(dsw) = subscriber
         .durable_alive
         .remove(&sub.path)
@@ -1118,7 +1250,7 @@ fn unsubscribe(
                     e.remove();
                 }
                 Some(s) => {
-                    if s.0.id == id && s.0.addr == addr {
+                    if s.0.id == id && s.0.conid == conid {
                         e.remove();
                     }
                 }
@@ -1170,11 +1302,6 @@ async fn hello_publisher(
 
 const PERIOD: Duration = Duration::from_secs(100);
 
-lazy_static! {
-    static ref BATCHES: Pool<Vec<(SubId, Event)>> = Pool::new(1000, 100000);
-    static ref DECODE_BATCHES: Pool<Vec<From>> = Pool::new(1000, 100000);
-}
-
 // This is the fast path for the common case where the batch contains
 // only updates. As of 2020-04-30, sending to an mpsc channel is
 // pretty slow, about 250ns, so we go to great lengths to avoid it.
@@ -1185,22 +1312,22 @@ async fn process_updates_batch(
 ) {
     for m in batch.drain(..) {
         if let From::Update(i, m) = m {
-            if let Some(sub) = subscriptions.get_mut(&i) {
-                for (sub_id, chan_id, c) in sub.streams.iter() {
+            if let Some(sub) = subscriptions.get(&i) {
+                for (chan_id, c) in sub.streams.0.iter() {
                     by_chan
                         .entry(*chan_id)
                         .or_insert_with(|| (c.clone(), BATCHES.take()))
                         .1
-                        .push((*sub_id, Event::Update(m.clone())))
+                        .push((sub.sub_id, Event::Update(m.clone())))
                 }
-                if sub.collect_last {
-                    *sub.last.lock() = Event::Update(m);
+                if let Some(last) = &sub.last {
+                    *last.lock() = Event::Update(m);
                 }
             }
         }
     }
     for (_, (mut c, batch)) in by_chan.drain() {
-        let _ = c.send(batch).await;
+        let _ = c.0.send(batch).await;
     }
 }
 
@@ -1212,21 +1339,21 @@ async fn process_batch(
     pending_writes: &mut HashMap<Id, VecDeque<oneshot::Sender<Value>>, FxBuildHasher>,
     con: &mut WriteChannel<ClientCtx>,
     subscriber: &Subscriber,
-    addr: SocketAddr,
+    conid: ConId,
 ) -> Result<()> {
     for m in batch.drain(..) {
         match m {
-            From::Update(i, m) => match subscriptions.get_mut(&i) {
+            From::Update(i, m) => match subscriptions.get(&i) {
                 Some(sub) => {
-                    for (sub_id, chan_id, c) in sub.streams.iter_mut() {
+                    for (chan_id, c) in sub.streams.0.iter() {
                         by_chan
                             .entry(*chan_id)
                             .or_insert_with(|| (c.clone(), BATCHES.take()))
                             .1
-                            .push((*sub_id, Event::Update(m.clone())));
+                            .push((sub.sub_id, Event::Update(m.clone())));
                     }
-                    if sub.collect_last {
-                        *sub.last.lock() = Event::Update(m);
+                    if let Some(last) = &sub.last {
+                        *last.lock() = Event::Update(m);
                     }
                 }
                 None => con.queue_send(&To::Unsubscribe(i))?,
@@ -1257,18 +1384,17 @@ async fn process_batch(
             From::Unsubscribed(id) => {
                 if let Some(s) = subscriptions.remove(&id) {
                     let mut t = subscriber.0.lock();
-                    unsubscribe(&mut *t, by_chan, s, id, addr);
+                    unsubscribe(&mut *t, by_chan, s, id, conid);
                 }
             }
             From::Subscribed(p, id, m) => match pending.remove(&p) {
                 None => con.queue_send(&To::Unsubscribe(id))?,
                 Some(req) => {
-                    let sub_id = SubId::new();
-                    let last = Arc::new(Mutex::new(Event::Update(m)));
+                    let last = TArc::new(Mutex::new(Event::Update(m)));
                     let s = Ok(Val(Arc::new(ValInner {
-                        sub_id,
+                        sub_id: req.sub_id,
                         id,
-                        addr,
+                        conid,
                         connection: req.con,
                         last: last.clone(),
                     })));
@@ -1279,9 +1405,9 @@ async fn process_batch(
                                 id,
                                 Sub {
                                     path: req.path,
-                                    last,
-                                    streams: Vec::new(),
-                                    collect_last: true,
+                                    sub_id: req.sub_id,
+                                    last: Some(last),
+                                    streams: Streams::new(),
                                 },
                             );
                         }
@@ -1291,7 +1417,7 @@ async fn process_batch(
         }
     }
     for (_, (mut c, batch)) in by_chan.drain() {
-        let _ = c.send(batch).await;
+        let _ = c.0.send(batch).await;
     }
     Ok(())
 }
@@ -1349,6 +1475,7 @@ async fn connection(
         HashMap::with_hasher(FxBuildHasher::default());
     let mut msg_recvd = false;
     let soc = time::timeout(PERIOD, TcpStream::connect(addr)).await??;
+    let conid = ConId::new();
     soc.set_nodelay(true)?;
     let mut con = Channel::new(soc);
     hello_publisher(&mut con, &auth, &target_spn).await?;
@@ -1414,34 +1541,34 @@ async fn connection(
                             }
                             ToCon::Stream { id, sub_id, mut tx, flags } => {
                                 if let Some(sub) = subscriptions.get_mut(&id) {
-                                    if flags.contains(UpdatesFlags::STOP_COLLECTING_LAST) {
-                                        sub.collect_last = false;
-                                    }
-                                    sub.streams.retain(|(_, _, c)| {
-                                        if c.is_closed() {
-                                            by_receiver.remove(&ChanWrap(c.clone()));
-                                            false
-                                        } else {
-                                            true
+                                    let mut already_have = false;
+                                    for (_, c) in sub.streams.0.iter() {
+                                        if tx.same_receiver(&c.0) {
+                                            already_have = true;
                                         }
-                                    });
+                                        if c.0.is_closed() {
+                                            by_receiver.remove(&c);
+                                        }
+                                    }
                                     if flags.contains(UpdatesFlags::BEGIN_WITH_LAST) {
-                                        let m = sub.last.lock().clone();
-                                        let mut b = BATCHES.take();
-                                        b.push((sub_id, m));
-                                        match tx.send(b).await {
-                                            Err(_) => continue,
-                                            Ok(()) => ()
+                                        if let Some(last) = &sub.last {
+                                            let m = last.lock().clone();
+                                            let mut b = BATCHES.take();
+                                            b.push((sub_id, m));
+                                            match tx.send(b).await {
+                                                Err(_) => continue,
+                                                Ok(()) => ()
+                                            }
                                         }
                                     }
-                                    let already_have =
-                                        sub.streams
-                                            .iter()
-                                            .any(|(_, _, s)| tx.same_receiver(s));
+                                    if flags.contains(UpdatesFlags::STOP_COLLECTING_LAST) {
+                                        sub.last = None;
+                                    }
                                     if !already_have {
-                                        let id = by_receiver.entry(ChanWrap(tx.clone()))
+                                        let tx = ChanWrap(tx);
+                                        let id = by_receiver.entry(tx.clone())
                                             .or_insert_with(ChanId::new);
-                                        sub.streams.push((sub_id, *id, tx));
+                                        sub.streams = sub.streams.add(*id, tx);
                                     }
                                 }
                             }
@@ -1487,7 +1614,7 @@ async fn connection(
                             &mut pending_writes,
                             &mut write_con,
                             &subscriber,
-                            addr).await);
+                            conid).await);
                         try_cf!(try_flush(&mut write_con));
                         if subscriptions.is_empty() && pending.is_empty() {
                             let mut inner = subscriber.0.lock();
@@ -1514,7 +1641,7 @@ async fn connection(
             &mut pending_writes,
             &mut write_con,
             &subscriber,
-            addr,
+            conid,
         )
         .await;
         for (_, req) in pending {
